# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Classes and functions for handling the ListSessionGroups API call."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import operator
import re

import six

from google.protobuf import struct_pb2

from tensorboard.plugins.hparams import api_pb2
from tensorboard.plugins.hparams import error
from tensorboard.plugins.hparams import metadata
from tensorboard.plugins.hparams import metrics


class Handler(object):
  """Handles a ListSessionGroups request."""

  def __init__(self, context, request):
    """Constructor.

    Args:
      context: A backend_context.Context instance.
      request: A ListSessionGroupsRequest protobuf.
    """
    self._context = context
    self._request = request
    self._extractors = _create_extractors(request.col_params)
    self._filters = _create_filters(request.col_params, self._extractors)
    # Since an context.experiment() call may search through all the runs, we
    # cache it here.
    self._experiment = context.experiment()

  def run(self):
    """Handles the request specified on construction.

    Returns:
      A ListSessionGroupsResponse object.

    """
    session_groups = self._build_session_groups()
    session_groups = self._filter(session_groups)
    self._sort(session_groups)
    return self._create_response(session_groups)

  def _build_session_groups(self):
    """Returns a list of SessionGroups protobuffers from the summary data."""

    # Algorithm: We keep a dict 'groups_by_name' mapping a SessionGroup name
    # (str) to a SessionGroup protobuffer. We traverse the runs associated with
    # the plugin--each representing a single session. We form a Session
    # protobuffer from each run and add it to the relevant SessionGroup object
    # in the 'groups_by_name' dict. We create the SessionGroup object, if this
    # is the first session of that group we encounter.
    groups_by_name = {}
    run_to_tag_to_content = self._context.multiplexer.PluginRunToTagToContent(
        metadata.PLUGIN_NAME)
    for (run, tag_to_content) in six.iteritems(run_to_tag_to_content):
      if metadata.SESSION_START_INFO_TAG not in tag_to_content:
        continue
      start_info = metadata.parse_session_start_info_plugin_data(
          tag_to_content[metadata.SESSION_START_INFO_TAG])
      end_info = None
      if metadata.SESSION_END_INFO_TAG in tag_to_content:
        end_info = metadata.parse_session_end_info_plugin_data(
            tag_to_content[metadata.SESSION_END_INFO_TAG])
      session = self._build_session(run, start_info, end_info)
      if session.status in self._request.allowed_statuses:
        self._add_session(session, start_info, groups_by_name)

    # Compute the session group's aggregated metrics for each group.
    groups = groups_by_name.values()
    for group in groups:
      # We sort the sessions in a group so that the order is deterministic.
      group.sessions.sort(key=operator.attrgetter('name'))
      self._aggregate_metrics(group)
    return groups

  def _add_session(self, session, start_info, groups_by_name):
    """Adds a new Session protobuffer to the 'groups_by_name' dictionary.

    Called by _build_session_groups when we encounter a new session. Creates
    the Session protobuffer and adds it to the relevant group in the
    'groups_by_name' dict. Creates the session group if this is the first time
    we encounter it.

    Args:
      session: api_pb2.Session. The session to add.
      start_info: The SessionStartInfo protobuffer associated with the session.
      groups_by_name: A str to SessionGroup protobuffer dict. Representing the
        session groups and sessions found so far.
    """
    # If the group_name is empty, this session's group contains only
    # this session. Use the session name for the group name since session
    # names are unique.
    group_name = start_info.group_name or session.name
    if group_name in groups_by_name:
      groups_by_name[group_name].sessions.extend([session])
    else:
      # Create the group and add the session as the first one.
      group = api_pb2.SessionGroup(
          name=group_name,
          sessions=[session],
          monitor_url=start_info.monitor_url)
      # Copy hparams from the first session (all sessions should have the same
      # hyperparameter values) into result.
      # There doesn't seem to be a way to initialize a protobuffer map in the
      # constructor.
      for (key, value) in six.iteritems(start_info.hparams):
        group.hparams[key].CopyFrom(value)
      groups_by_name[group_name] = group

  def _build_session(self, name, start_info, end_info):
    """Builds a session object."""

    assert start_info is not None
    result = api_pb2.Session(
        name=name,
        start_time_secs=start_info.start_time_secs,
        model_uri=start_info.model_uri,
        metric_values=self._build_session_metric_values(name),
        monitor_url=start_info.monitor_url)
    if end_info is not None:
      result.status = end_info.status
      result.end_time_secs = end_info.end_time_secs
    return result

  def _build_session_metric_values(self, session_name):
    """Builds the session metric values."""

    # result is a list of api_pb2.MetricValue instances.
    result = []
    metric_infos = self._experiment.metric_infos
    for metric_info in metric_infos:
      metric_name = metric_info.name
      try:
        metric_eval = metrics.last_metric_eval(
            self._context.multiplexer,
            session_name,
            metric_name)
      except KeyError:
        # It's ok if we don't find the metric in the session.
        # We skip it here. For filtering and sorting purposes its value is None.
        continue

      # metric_eval is a 3-tuple of the form [wall_time, step, value]
      result.append(api_pb2.MetricValue(name=metric_name,
                                        wall_time_secs=metric_eval[0],
                                        training_step=metric_eval[1],
                                        value=metric_eval[2]))
    return result

  def _aggregate_metrics(self, session_group):
    """Sets the metrics of the group based on aggregation_type."""

    if (self._request.aggregation_type == api_pb2.AGGREGATION_AVG or
        self._request.aggregation_type == api_pb2.AGGREGATION_UNSET):
      _set_avg_session_metrics(session_group)
    elif self._request.aggregation_type == api_pb2.AGGREGATION_MEDIAN:
      _set_median_session_metrics(session_group,
                                  self._request.aggregation_metric)
    elif self._request.aggregation_type == api_pb2.AGGREGATION_MIN:
      _set_extremum_session_metrics(session_group,
                                    self._request.aggregation_metric,
                                    min)
    elif self._request.aggregation_type == api_pb2.AGGREGATION_MAX:
      _set_extremum_session_metrics(session_group,
                                    self._request.aggregation_metric,
                                    max)
    else:
      raise error.HParamsError('Unknown aggregation_type in request: %s' %
                               self._request.aggregation_type)

  def _filter(self, session_groups):
    return [sg for sg in session_groups if self._passes_all_filters(sg)]

  def _passes_all_filters(self, session_group):
    return all(filter_fn(session_group) for filter_fn in self._filters)

  def _sort(self, session_groups):
    """Sorts 'session_groups' in place according to _request.col_params."""

    # Sort by session_group name so we have a deterministic order.
    session_groups.sort(key=operator.attrgetter('name'))
    # Sort by lexicographical order of the _request.col_params whose order
    # is not ORDER_UNSPECIFIED. The first such column is the primary sorting
    # key, the second is the secondary sorting key, etc. To achieve that we
    # need to iterate on these columns in reverse order (thus the primary key
    # is the key used in the last sort).
    for col_param, extractor in reversed(list(zip(self._request.col_params,
                                                  self._extractors))):
      if col_param.order == api_pb2.ORDER_UNSPECIFIED:
        continue
      if col_param.order == api_pb2.ORDER_ASC:
        session_groups.sort(
            key=_create_key_func(
                extractor,
                none_is_largest=not col_param.missing_values_first))
      elif col_param.order == api_pb2.ORDER_DESC:
        session_groups.sort(
            key=_create_key_func(
                extractor,
                none_is_largest=col_param.missing_values_first),
            reverse=True)
      else:
        raise error.HParamsError('Unknown col_param.order given: %s' %
                                 col_param)

  def _create_response(self, session_groups):
    return api_pb2.ListSessionGroupsResponse(
        session_groups=session_groups[
            self._request.start_index:
            self._request.start_index+self._request.slice_size],
        total_size=len(session_groups))


def _create_key_func(extractor, none_is_largest):
  """Returns a key_func to be used in list.sort().

  Returns a key_func to be used in list.sort() that sorts session groups
  by the value extracted by extractor. 'None' extracted values will either
  be considered largest or smallest as specified by the "none_is_largest"
  boolean parameter.

  Args:
    extractor: An extractor function that extract the key from the session
      group.
    none_is_largest: bool. If true treats 'None's as largest; otherwise
      smallest.

  """
  if none_is_largest:
    def key_func_none_is_largest(session_group):
      value = extractor(session_group)
      return (value is None, value)
    return key_func_none_is_largest
  def key_func_none_is_smallest(session_group):
    value = extractor(session_group)
    return (value is not None, value)
  return key_func_none_is_smallest


# Extractors. An extractor is a function that extracts some property (a metric
# or a hyperparameter) from a SessionGroup instance.
def _create_extractors(col_params):
  """Creates extractors to extract properties corresponding to 'col_params'.

  Args:
    col_params: List of ListSessionGroupsRequest.ColParam protobufs.
  Returns:
    A list of extractor functions. The ith element in the
    returned list extracts the column corresponding to the ith element of
    _request.col_params
  """
  result = []
  for col_param in col_params:
    result.append(_create_extractor(col_param))
  return result


def _create_extractor(col_param):
  if col_param.HasField('metric'):
    return _create_metric_extractor(col_param.metric)
  elif col_param.HasField('hparam'):
    return _create_hparam_extractor(col_param.hparam)
  else:
    raise error.HParamsError(
        'Got ColParam with both "metric" and "hparam" fields unset: %s' %
        col_param)


def _create_metric_extractor(metric_name):
  """Returns function that extracts a metric from a session group or a session.

  Args:
    metric_name: tensorboard.hparams.MetricName protobuffer. Identifies the
    metric to extract from the session group.
  Returns:
    A function that takes a tensorboard.hparams.SessionGroup or
    tensorborad.hparams.Session protobuffer and returns the value of the metric
    identified by 'metric_name' or None if the value doesn't exist.
  """
  def extractor_fn(session_or_group):
    metric_value = _find_metric_value(session_or_group,
                                      metric_name)
    return metric_value.value if metric_value else None

  return extractor_fn


def _find_metric_value(session_or_group, metric_name):
  """Returns the metric_value for a given metric in a session or session group.

  Args:
    session_or_group: A Session protobuffer or SessionGroup protobuffer.
    metric_name: A MetricName protobuffer. The metric to search for.
  Returns:
    A MetricValue protobuffer representing the value of the given metric or
    None if no such metric was found in session_or_group.
  """
  # Note: We can speed this up by converting the metric_values field
  # to a dictionary on initialization, to avoid a linear search here. We'll
  # need to wrap the SessionGroup and Session protos in a python object for
  # that.
  for metric_value in session_or_group.metric_values:
    if (metric_value.name.tag == metric_name.tag and
        metric_value.name.group == metric_name.group):
      return metric_value


def _create_hparam_extractor(hparam_name):
  """Returns an extractor function that extracts an hparam from a session group.

  Args:
    hparam_name: str. Identies the hparam to extract from the session group.
  Returns:
    A function that takes a tensorboard.hparams.SessionGroup protobuffer and
    returns the value, as a native Python object, of the hparam identified by
    'hparam_name'.
  """
  def extractor_fn(session_group):
    if hparam_name in session_group.hparams:
      return _value_to_python(session_group.hparams[hparam_name])
    return None

  return extractor_fn


# Filters. A filter is a boolean function that takes a session group and returns
# True if it should be included in the result. Currently, Filters are functions
# of a single column value extracted from the session group with a given
# extractor specified in the construction of the filter.
def _create_filters(col_params, extractors):
  """Creates filters for the given col_params.

  Args:
    col_params: List of ListSessionGroupsRequest.ColParam protobufs.
    extractors: list of extractor functions of the same length as col_params.
      Each element should extract the column described by the corresponding
      element of col_params.
  Returns:
    A list of filter functions. Each corresponding to a single
    col_params.filter oneof field of _request
  """
  result = []
  for col_param, extractor in zip(col_params, extractors):
    a_filter = _create_filter(col_param, extractor)
    if a_filter:
      result.append(a_filter)
  return result


def _create_filter(col_param, extractor):
  """Creates a filter for the given col_param and extractor.

  Args:
    col_param: A tensorboard.hparams.ColParams object identifying the column
      and describing the filter to apply.
    extractor: A function that extract the column value identified by
      'col_param' from a tensorboard.hparams.SessionGroup protobuffer.
  Returns:
    A boolean function taking a tensorboard.hparams.SessionGroup protobuffer
    returning True if the session group passes the filter described by
    'col_param'. If col_param does not specify a filter (i.e. any session
    group passes) returns None.
  """
  include_missing_values = not col_param.exclude_missing_values
  if col_param.HasField('filter_regexp'):
    value_filter_fn = _create_regexp_filter(col_param.filter_regexp)
  elif col_param.HasField('filter_interval'):
    value_filter_fn = _create_interval_filter(col_param.filter_interval)
  elif col_param.HasField('filter_discrete'):
    value_filter_fn = _create_discrete_set_filter(col_param.filter_discrete)
  elif include_missing_values:
    # No 'filter' field and include_missing_values is True.
    # Thus, the resulting filter always returns True, so to optimize for this
    # common case we do not include it in the list of filters to check.
    return None
  else:
    value_filter_fn = lambda _: True

  def filter_fn(session_group):
    value = extractor(session_group)
    if value is None:
      return include_missing_values
    return value_filter_fn(value)

  return filter_fn


def _create_regexp_filter(regex):
  """Returns a boolean function that filters strings based on a regular exp.

  Args:
    regex: A string describing the regexp to use.
  Returns:
    A function taking a string and returns True if any of its substrings
    matches regex.
  """
  # Warning: Note that python's regex library allows inputs that take
  # exponential time. Time-limiting it is difficult. When we move to
  # a true multi-tenant tensorboard server, the regexp implementation here
  # would need to be replaced by something more secure.
  compiled_regex = re.compile(regex)
  def filter_fn(value):
    if not isinstance(value, six.string_types):
      raise error.HParamsError(
          'Cannot use a regexp filter for a value of type %s. Value: %s' %
          (type(value), value))
    return re.search(compiled_regex, value) is not None

  return filter_fn


def _create_interval_filter(interval):
  """Returns a function that checkes whether a number belongs to an interval.

  Args:
    interval: A tensorboard.hparams.Interval protobuf describing the interval.
  Returns:
    A function taking a number (a float or an object of a type in
    six.integer_types) that returns True if the number belongs to (the closed)
    'interval'.
  """
  def filter_fn(value):
    if (not isinstance(value, six.integer_types) and
        not isinstance(value, float)):
      raise error.HParamsError(
          'Cannot use an interval filter for a value of type: %s, Value: %s' %
          (type(value), value))
    return interval.min_value <= value and value <= interval.max_value

  return filter_fn


def _create_discrete_set_filter(discrete_set):
  """Returns a function that checks whether a value belongs to a set.

  Args:
    discrete_set: A list of objects representing the set.
  Returns:
    A function taking an object and returns True if its in the set. Membership
    is tested using the Python 'in' operator (thus, equality of distinct
    objects is computed using the '==' operator).
  """
  def filter_fn(value):
    return value in discrete_set

  return filter_fn


def _value_to_python(value):
  """Converts a google.protobuf.Value to a native Python object."""

  assert isinstance(value, struct_pb2.Value)
  field = value.WhichOneof('kind')
  if field == 'number_value':
    return value.number_value
  elif field == 'string_value':
    return value.string_value
  elif field == 'bool_value':
    return value.bool_value
  else:
    raise ValueError('Unknown struct_pb2.Value oneof field set: %s' % field)


# As protobuffers are mutable we can't use MetricName directly as a dict's key.
# Instead, we duplicate MetricName as a collections.namedtuple. Another
# alternative would be to wrap the MetricName protocol buffer class in an
# immutable class that defines equality and __hash__ methods based on the text
# representation of the protocol buffer. This is more complex, but won't
# require modification if we ever add fields to MetricName.
_MetricIdentifier = collections.namedtuple('_MetricIdentifier', 'group tag')


class _MetricStats(object):
  """A simple class to hold metric stats used in calculating metric averages.

  Used in _set_avg_session_metrics(). See the comments in that function
  for more details.

  Attributes:
    total: int. The sum of the metric measurements seen so far.
    count: int. The number of largest-step measuremens seen so far.
    total_step: int. The sum of the steps at which the measurements were taken
    total_wall_time_secs: float. The sum of the wall_time_secs at
        which the measurements were taken.
  """
  # We use slots here to catch typos in attributes earlier. Note that this makes
  # this class incompatible with 'pickle'.
  __slots__ = [
      'total',
      'count',
      'total_step',
      'total_wall_time_secs',
  ]

  def __init__(self):
    self.total = 0
    self.count = 0
    self.total_step = 0
    self.total_wall_time_secs = 0.0


def _set_avg_session_metrics(session_group):
  """Sets the metrics for the group to be the average of its sessions.

  The resulting session group metrics consist of the union of metrics across
  the group's sessions. The value of each session group metric is the average
  of that metric values across the sessions in the group. The 'step' and
  'wall_time_secs' fields of the resulting MetricValue field in the session
  group are populated with the corresponding averages (truncated for 'step')
  as well.

  Args:
    session_group: A SessionGroup protobuffer.
  """
  assert session_group.sessions, 'SessionGroup cannot be empty.'
  # Algorithm: Iterate over all (session, metric) pairs and maintain a
  # dict from _MetricIdentifier to _MetricStats objects.
  # Then use the final dict state to compute the average for each metric.
  metric_stats = collections.defaultdict(_MetricStats)
  for session in session_group.sessions:
    for metric_value in session.metric_values:
      metric_name = _MetricIdentifier(group=metric_value.name.group,
                                      tag=metric_value.name.tag)
      stats = metric_stats[metric_name]
      stats.total += metric_value.value
      stats.count += 1
      stats.total_step += metric_value.training_step
      stats.total_wall_time_secs += metric_value.wall_time_secs

  del session_group.metric_values[:]
  for (metric_name, stats) in six.iteritems(metric_stats):
    session_group.metric_values.add(
        name=api_pb2.MetricName(group=metric_name.group, tag=metric_name.tag),
        value=float(stats.total)/float(stats.count),
        training_step=stats.total_step // stats.count,
        wall_time_secs=stats.total_wall_time_secs / stats.count)


# A namedtuple to hold a session's metric value.
# 'session_index' is the index of the session in its group.
_Measurement = collections.namedtuple('_Measurement', ['metric_value',
                                                       'session_index'])


def _set_median_session_metrics(session_group, aggregation_metric):
  """Sets the metrics for session_group to those of its "median session".

  The median session is the session in session_group with the median value
  of the metric given by 'aggregation_metric'. The median is taken over the
  subset of sessions in the group whose 'aggregation_metric' was measured
  at the largest training step among the sessions in the group.

  Args:
    session_group: A SessionGroup protobuffer.
    aggregation_metric: A MetricName protobuffer.
  """
  measurements = sorted(_measurements(session_group, aggregation_metric),
                        key=operator.attrgetter('metric_value.value'))
  median_session = measurements[(len(measurements) - 1) // 2].session_index
  del session_group.metric_values[:]
  session_group.metric_values.MergeFrom(
      session_group.sessions[median_session].metric_values)


def _set_extremum_session_metrics(session_group, aggregation_metric,
                                  extremum_fn):
  """Sets the metrics for session_group to those of its "extremum session".

  The extremum session is the session in session_group with the extremum value
  of the metric given by 'aggregation_metric'. The extremum is taken over the
  subset of sessions in the group whose 'aggregation_metric' was measured
  at the largest training step among the sessions in the group.

  Args:
    session_group: A SessionGroup protobuffer.
    aggregation_metric: A MetricName protobuffer.
    extremum_fn: callable. Must be either 'min' or 'max'. Determines the type of
      extremum to compute.
  """
  measurements = _measurements(session_group, aggregation_metric)
  ext_session = extremum_fn(
      measurements,
      key=operator.attrgetter('metric_value.value')).session_index
  del session_group.metric_values[:]
  session_group.metric_values.MergeFrom(
      session_group.sessions[ext_session].metric_values)


def _measurements(session_group, metric_name):
  """A generator for the values of the metric across the sessions in the group.

  Args:
    session_group: A SessionGroup protobuffer.
    metric_name: A MetricName protobuffer.
  Yields:
    The next metric value wrapped in a _Measurement instance.
  """
  for session_index, session in enumerate(session_group.sessions):
    metric_value = _find_metric_value(session, metric_name)
    if not metric_value:
      continue
    yield _Measurement(metric_value, session_index)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The TensorBoard Graphs plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import six
from werkzeug import wrappers

from tensorboard.backend import http_util
from tensorboard.backend import process_graph
from tensorboard.backend.event_processing import plugin_event_accumulator as event_accumulator  # pylint: disable=line-too-long
from tensorboard.compat.proto import config_pb2
from tensorboard.compat.proto import graph_pb2
from tensorboard.plugins import base_plugin
from tensorboard.plugins.graph import graph_util
from tensorboard.plugins.graph import keras_util
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

_PLUGIN_PREFIX_ROUTE = 'graphs'

# The Summary API is implemented in TensorFlow because it uses TensorFlow internal APIs.
# As a result, this SummaryMetadata is a bit unconventional and uses non-public
# hardcoded name as the plugin name. Please refer to link below for the summary ops.
# https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L757
_PLUGIN_NAME_RUN_METADATA = 'graph_run_metadata'
# https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L788
_PLUGIN_NAME_RUN_METADATA_WITH_GRAPH = 'graph_run_metadata_graph'
# https://github.com/tensorflow/tensorflow/blob/565952cc2f17fdfd995e25171cf07be0f6f06180/tensorflow/python/ops/summary_ops_v2.py#L825
_PLUGIN_NAME_KERAS_MODEL = 'graph_keras_model'


class GraphsPlugin(base_plugin.TBPlugin):
  """Graphs Plugin for TensorBoard."""

  plugin_name = _PLUGIN_PREFIX_ROUTE

  def __init__(self, context):
    """Instantiates GraphsPlugin via TensorBoard core.

    Args:
      context: A base_plugin.TBContext instance.
    """
    self._multiplexer = context.multiplexer

  def get_plugin_apps(self):
    return {
        '/graph': self.graph_route,
        '/info': self.info_route,
        '/run_metadata': self.run_metadata_route,
    }

  def is_active(self):
    """The graphs plugin is active iff any run has a graph."""
    return bool(self._multiplexer and self.info_impl())

  def frontend_metadata(self):
    return super(GraphsPlugin, self).frontend_metadata()._replace(
        element_name='tf-graph-dashboard',
        # TODO(@chihuahua): Reconcile this setting with Health Pills.
        disable_reload=True,
    )

  def info_impl(self):
    """Returns a dict of all runs and tags and their data availabilities."""
    result = {}
    def add_row_item(run, tag=None):
      run_item = result.setdefault(run, {
          'run': run,
          'tags': {},
          # A run-wide GraphDef of ops.
          'run_graph': False})

      tag_item = None
      if tag:
        tag_item = run_item.get('tags').setdefault(tag, {
            'tag': tag,
            'conceptual_graph': False,
            # A tagged GraphDef of ops.
            'op_graph': False,
            'profile': False})
      return (run_item, tag_item)

    mapping = self._multiplexer.PluginRunToTagToContent(
        _PLUGIN_NAME_RUN_METADATA_WITH_GRAPH)
    for run_name, tag_to_content in six.iteritems(mapping):
      for (tag, content) in six.iteritems(tag_to_content):
        # The Summary op is defined in TensorFlow and does not use a stringified proto
        # as a content of plugin data. It contains single string that denotes a version.
        # https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L789-L790
        if content != b'1':
          logger.warn('Ignoring unrecognizable version of RunMetadata.')
          continue
        (_, tag_item) = add_row_item(run_name, tag)
        tag_item['op_graph'] = True

    # Tensors associated with plugin name _PLUGIN_NAME_RUN_METADATA contain
    # both op graph and profile information.
    mapping = self._multiplexer.PluginRunToTagToContent(
        _PLUGIN_NAME_RUN_METADATA)
    for run_name, tag_to_content in six.iteritems(mapping):
      for (tag, content) in six.iteritems(tag_to_content):
        if content != b'1':
          logger.warn('Ignoring unrecognizable version of RunMetadata.')
          continue
        (_, tag_item) = add_row_item(run_name, tag)
        tag_item['profile'] = True
        tag_item['op_graph'] = True

    # Tensors associated with plugin name _PLUGIN_NAME_KERAS_MODEL contain
    # serialized Keras model in JSON format.
    mapping = self._multiplexer.PluginRunToTagToContent(
        _PLUGIN_NAME_KERAS_MODEL)
    for run_name, tag_to_content in six.iteritems(mapping):
      for (tag, content) in six.iteritems(tag_to_content):
        if content != b'1':
          logger.warn('Ignoring unrecognizable version of RunMetadata.')
          continue
        (_, tag_item) = add_row_item(run_name, tag)
        tag_item['conceptual_graph'] = True

    for (run_name, run_data) in six.iteritems(self._multiplexer.Runs()):
      if run_data.get(event_accumulator.GRAPH):
        (run_item, _) = add_row_item(run_name, None)
        run_item['run_graph'] = True

    for (run_name, run_data) in six.iteritems(self._multiplexer.Runs()):
      if event_accumulator.RUN_METADATA in run_data:
        for tag in run_data[event_accumulator.RUN_METADATA]:
          (_, tag_item) = add_row_item(run_name, tag)
          tag_item['profile'] = True

    return result

  def graph_impl(self, run, tag, is_conceptual, limit_attr_size=None, large_attrs_key=None):
    """Result of the form `(body, mime_type)`, or `None` if no graph exists."""
    if is_conceptual:
      tensor_events = self._multiplexer.Tensors(run, tag)
      # Take the first event if there are multiple events written from different
      # steps.
      keras_model_config = json.loads(tensor_events[0].tensor_proto.string_val[0])
      graph = keras_util.keras_model_to_graph_def(keras_model_config)
    elif tag:
      tensor_events = self._multiplexer.Tensors(run, tag)
      # Take the first event if there are multiple events written from different
      # steps.
      run_metadata = config_pb2.RunMetadata.FromString(
          tensor_events[0].tensor_proto.string_val[0])
      graph = graph_pb2.GraphDef()

      for func_graph in run_metadata.function_graphs:
        graph_util.combine_graph_defs(graph, func_graph.pre_optimization_graph)
    else:
      graph = self._multiplexer.Graph(run)

    # This next line might raise a ValueError if the limit parameters
    # are invalid (size is negative, size present but key absent, etc.).
    process_graph.prepare_graph_for_ui(graph, limit_attr_size, large_attrs_key)
    return (str(graph), 'text/x-protobuf')  # pbtxt

  def run_metadata_impl(self, run, tag):
    """Result of the form `(body, mime_type)`, or `None` if no data exists."""
    try:
      run_metadata = self._multiplexer.RunMetadata(run, tag)
    except ValueError:
      # TODO(stephanwlee): Should include whether FE is fetching for v1 or v2 RunMetadata
      # so we can remove this try/except.
      tensor_events = self._multiplexer.Tensors(run, tag)
      if tensor_events is None:
        return None
      # Take the first event if there are multiple events written from different
      # steps.
      run_metadata = config_pb2.RunMetadata.FromString(
          tensor_events[0].tensor_proto.string_val[0])
    if run_metadata is None:
      return None
    return (str(run_metadata), 'text/x-protobuf')  # pbtxt

  @wrappers.Request.application
  def info_route(self, request):
    info = self.info_impl()
    return http_util.Respond(request, info, 'application/json')

  @wrappers.Request.application
  def graph_route(self, request):
    """Given a single run, return the graph definition in protobuf format."""
    run = request.args.get('run')
    tag = request.args.get('tag', '')
    conceptual_arg = request.args.get('conceptual', False)
    is_conceptual = True if conceptual_arg == 'true' else False

    if run is None:
      return http_util.Respond(
          request, 'query parameter "run" is required', 'text/plain', 400)

    limit_attr_size = request.args.get('limit_attr_size', None)
    if limit_attr_size is not None:
      try:
        limit_attr_size = int(limit_attr_size)
      except ValueError:
        return http_util.Respond(
            request, 'query parameter `limit_attr_size` must be an integer',
            'text/plain', 400)

    large_attrs_key = request.args.get('large_attrs_key', None)

    try:
      result = self.graph_impl(run, tag, is_conceptual, limit_attr_size, large_attrs_key)
    except ValueError as e:
      return http_util.Respond(request, e.message, 'text/plain', code=400)
    else:
      if result is not None:
        (body, mime_type) = result  # pylint: disable=unpacking-non-sequence
        return http_util.Respond(request, body, mime_type)
      else:
        return http_util.Respond(request, '404 Not Found', 'text/plain',
                                 code=404)

  @wrappers.Request.application
  def run_metadata_route(self, request):
    """Given a tag and a run, return the session.run() metadata."""
    tag = request.args.get('tag')
    run = request.args.get('run')
    if tag is None:
      return http_util.Respond(
          request, 'query parameter "tag" is required', 'text/plain', 400)
    if run is None:
      return http_util.Respond(
          request, 'query parameter "run" is required', 'text/plain', 400)
    result = self.run_metadata_impl(run, tag)
    if result is not None:
      (body, mime_type) = result  # pylint: disable=unpacking-non-sequence
      return http_util.Respond(request, body, mime_type)
    else:
      return http_util.Respond(request, '404 Not Found', 'text/plain',
                               code=404)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Mesh summaries and TensorFlow operations to create them."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import tensorflow as tf

from tensorboard.plugins.mesh import metadata
from tensorboard.plugins.mesh import plugin_data_pb2

PLUGIN_NAME = 'mesh'


def _get_tensor_summary(
    name, display_name, description, tensor, content_type, json_config,
    collections):
  """Creates a tensor summary with summary metadata.

  Args:
    name: Uniquely identifiable name of the summary op. Could be replaced by
      combination of name and type to make it unique even outside of this
      summary.
    display_name: Will be used as the display name in TensorBoard.
      Defaults to `tag`.
    description: A longform readable description of the summary data. Markdown
      is supported.
    tensor: Tensor to display in summary.
    content_type: Type of content inside the Tensor.
    json_config: A string, JSON-serialized dictionary of ThreeJS classes
      configuration.
    collections: List of collections to add this summary to.

  Returns:
    Tensor summary with metadata.
  """
  tensor = tf.convert_to_tensor(value=tensor)
  tensor_metadata = metadata.create_summary_metadata(
      name,
      display_name,
      content_type,
      tensor.shape.as_list(),
      description,
      json_config=json_config)
  tensor_summary = tf.summary.tensor_summary(
      metadata.get_instance_name(name, content_type),
      tensor,
      summary_metadata=tensor_metadata,
      collections=collections)
  return tensor_summary


def _get_display_name(name, display_name):
  """Returns display_name from display_name and name."""
  if display_name is None:
    return name
  return display_name


def _get_json_config(config_dict):
  """Parses and returns JSON string from python dictionary."""
  json_config = '{}'
  if config_dict is not None:
    json_config = json.dumps(config_dict, sort_keys=True)
  return json_config


def op(name, vertices, faces=None, colors=None, display_name=None,
       description=None, collections=None, config_dict=None):
  """Creates a TensorFlow summary op for mesh rendering.

  Args:
    name: A name for this summary operation.
    vertices: Tensor of shape `[dim_1, ..., dim_n, 3]` representing the 3D
      coordinates of vertices.
    faces: Tensor of shape `[dim_1, ..., dim_n, 3]` containing indices of
      vertices within each triangle.
    colors: Tensor of shape `[dim_1, ..., dim_n, 3]` containing colors for each
      vertex.
    display_name: If set, will be used as the display name in TensorBoard.
      Defaults to `name`.
    description: A longform readable description of the summary data. Markdown
      is supported.
    collections: Which TensorFlow graph collections to add the summary op to.
      Defaults to `['summaries']`. Can usually be ignored.
    config_dict: Dictionary with ThreeJS classes names and configuration.

  Returns:
    Merged summary for mesh/point cloud representation.
  """
  display_name = _get_display_name(name, display_name)
  json_config = _get_json_config(config_dict)

  # All tensors representing a single mesh will be represented as separate
  # summaries internally. Those summaries will be regrouped on the client before
  # rendering.
  summaries = []
  tensors = [
      (vertices, plugin_data_pb2.MeshPluginData.VERTEX),
      (faces, plugin_data_pb2.MeshPluginData.FACE),
      (colors, plugin_data_pb2.MeshPluginData.COLOR)
  ]

  for tensor, content_type in tensors:
    if tensor is None:
      continue
    summaries.append(
        _get_tensor_summary(name, display_name, description, tensor,
                            content_type, json_config, collections))

  all_summaries = tf.summary.merge(
      summaries, collections=collections, name=name)
  return all_summaries


def pb(name,
       vertices,
       faces=None,
       colors=None,
       display_name=None,
       description=None,
       config_dict=None):
  """Create a mesh summary to save in pb format.

  Args:
    name: A name for this summary operation.
    vertices: numpy array of shape `[dim_1, ..., dim_n, 3]` representing the 3D
      coordinates of vertices.
    faces: numpy array of shape `[dim_1, ..., dim_n, 3]` containing indices of
      vertices within each triangle.
    colors: numpy array of shape `[dim_1, ..., dim_n, 3]` containing colors for
      each vertex.
    display_name: If set, will be used as the display name in TensorBoard.
      Defaults to `name`.
    description: A longform readable description of the summary data. Markdown
      is supported.
    config_dict: Dictionary with ThreeJS classes names and configuration.

  Returns:
    Instance of tf.Summary class.
  """
  display_name = _get_display_name(name, display_name)
  json_config = _get_json_config(config_dict)

  summaries = []
  tensors = [(vertices, plugin_data_pb2.MeshPluginData.VERTEX, tf.float32),
             (faces, plugin_data_pb2.MeshPluginData.FACE, tf.int32),
             (colors, plugin_data_pb2.MeshPluginData.COLOR, tf.uint8)]
  for tensor, content_type, data_type in tensors:
    if tensor is None:
      continue
    tensor_shape = tensor.shape
    tensor = tf.compat.v1.make_tensor_proto(tensor, dtype=data_type)
    summary_metadata = metadata.create_summary_metadata(
        name,
        display_name,
        content_type,
        tensor_shape,
        description,
        json_config=json_config)
    tag = metadata.get_instance_name(name, content_type)
    summaries.append((tag, summary_metadata, tensor))

  summary = tf.Summary()
  for tag, summary_metadata, tensor in summaries:
    tf_summary_metadata = tf.SummaryMetadata.FromString(
        summary_metadata.SerializeToString())
    summary.value.add(tag=tag, metadata=tf_summary_metadata, tensor=tensor)
  return summary



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorboard/plugins/mesh/plugin_data.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorboard/plugins/mesh/plugin_data.proto',
  package='tensorboard.mesh',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n*tensorboard/plugins/mesh/plugin_data.proto\x12\x10tensorboard.mesh\"\xd6\x01\n\x0eMeshPluginData\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x42\n\x0c\x63ontent_type\x18\x03 \x01(\x0e\x32,.tensorboard.mesh.MeshPluginData.ContentType\x12\x13\n\x0bjson_config\x18\x05 \x01(\t\x12\r\n\x05shape\x18\x06 \x03(\x05\"=\n\x0b\x43ontentType\x12\r\n\tUNDEFINED\x10\x00\x12\n\n\x06VERTEX\x10\x01\x12\x08\n\x04\x46\x41\x43\x45\x10\x02\x12\t\n\x05\x43OLOR\x10\x03\x62\x06proto3')
)



_MESHPLUGINDATA_CONTENTTYPE = _descriptor.EnumDescriptor(
  name='ContentType',
  full_name='tensorboard.mesh.MeshPluginData.ContentType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VERTEX', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FACE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COLOR', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=218,
  serialized_end=279,
)
_sym_db.RegisterEnumDescriptor(_MESHPLUGINDATA_CONTENTTYPE)


_MESHPLUGINDATA = _descriptor.Descriptor(
  name='MeshPluginData',
  full_name='tensorboard.mesh.MeshPluginData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='tensorboard.mesh.MeshPluginData.version', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='tensorboard.mesh.MeshPluginData.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='content_type', full_name='tensorboard.mesh.MeshPluginData.content_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='json_config', full_name='tensorboard.mesh.MeshPluginData.json_config', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='tensorboard.mesh.MeshPluginData.shape', index=4,
      number=6, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MESHPLUGINDATA_CONTENTTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=65,
  serialized_end=279,
)

_MESHPLUGINDATA.fields_by_name['content_type'].enum_type = _MESHPLUGINDATA_CONTENTTYPE
_MESHPLUGINDATA_CONTENTTYPE.containing_type = _MESHPLUGINDATA
DESCRIPTOR.message_types_by_name['MeshPluginData'] = _MESHPLUGINDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MeshPluginData = _reflection.GeneratedProtocolMessageType('MeshPluginData', (_message.Message,), dict(
  DESCRIPTOR = _MESHPLUGINDATA,
  __module__ = 'tensorboard.plugins.mesh.plugin_data_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.mesh.MeshPluginData)
  ))
_sym_db.RegisterMessage(MeshPluginData)


# @@protoc_insertion_point(module_scope)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility interfaces for TensorBoard.

This module provides logic for importing variations on the TensorFlow APIs, as
lazily loaded imports to help avoid circular dependency issues and defer the
search and loading of the module until necessary.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib as _importlib

import tensorboard.lazy as _lazy


@_lazy.lazy_load('tensorboard.compat.tf')
def tf():
  """Provide the root module of a TF-like API for use within TensorBoard.

  By default this is equivalent to `import tensorflow as tf`, but it can be used
  in combination with //tensorboard/compat:tensorflow (to fall back to a stub TF
  API implementation if the real one is not available) or with
  //tensorboard/compat:no_tensorflow (to force unconditional use of the stub).

  Returns:
    The root module of a TF-like API, if available.

  Raises:
    ImportError: if a TF-like API is not available.
  """
  try:
    from tensorboard.compat import notf  # pylint: disable=g-import-not-at-top
  except ImportError:
    try:
      import tensorflow  # pylint: disable=g-import-not-at-top
      return tensorflow
    except ImportError:
      pass
  from tensorboard.compat import tensorflow_stub  # pylint: disable=g-import-not-at-top
  return tensorflow_stub


@_lazy.lazy_load('tensorboard.compat.tf2')
def tf2():
  """Provide the root module of a TF-2.0 API for use within TensorBoard.

  Returns:
    The root module of a TF-2.0 API, if available.

  Raises:
    ImportError: if a TF-2.0 API is not available.
  """
  # Import the `tf` compat API from this file and check if it's already TF 2.0.
  if tf.__version__.startswith('2.'):
    return tf
  elif hasattr(tf, 'compat') and hasattr(tf.compat, 'v2'):
    # As a fallback, try `tensorflow.compat.v2` if it's defined.
    return tf.compat.v2
  raise ImportError('cannot import tensorflow 2.0 API')


# TODO(https://github.com/tensorflow/tensorboard/issues/1711): remove this
@_lazy.lazy_load('tensorboard.compat._pywrap_tensorflow')
def _pywrap_tensorflow():
  """Provide pywrap_tensorflow access in TensorBoard.

  pywrap_tensorflow cannot be accessed from tf.python.pywrap_tensorflow
  and needs to be imported using
  `from tensorflow.python import pywrap_tensorflow`. Therefore, we provide
  a separate accessor function for it here.

  NOTE: pywrap_tensorflow is not part of TensorFlow API and this
  dependency will go away soon.

  Returns:
    pywrap_tensorflow import, if available.

  Raises:
    ImportError: if we couldn't import pywrap_tensorflow.
  """
  try:
    from tensorboard.compat import notf  # pylint: disable=g-import-not-at-top
  except ImportError:
    try:
      from tensorflow.python import pywrap_tensorflow  # pylint: disable=g-import-not-at-top
      return pywrap_tensorflow
    except ImportError:
      pass
  from tensorboard.compat.tensorflow_stub import pywrap_tensorflow  # pylint: disable=g-import-not-at-top
  return pywrap_tensorflow
